

from typing import Iterable
from .UIModl import Content


class AbcArgs(object):
    """
    docstring
    """
    pass

    def __init__(self, title, default=None, required=True) -> None:
        self.title = title
        self.default = default
        self.required = required
        self.var_name: str = None

    def set_var_name(self, var_name) -> 'AbcArgs':
        self.var_name = var_name
        return self

    def to_content(self) -> Content:
        raise NotImplementedError


class ColumnSelect(AbcArgs):

    def __init__(self, title, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

    def to_content(self) -> Content:
        ct = Content(self.title, 'select', self.var_name,
                     source='data_columns', defaultValue=self.default)
        return ct


class Input(AbcArgs):

    def __init__(self, title, placeholder=None, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)
        self.placeholder = placeholder

    def to_content(self) -> Content:
        ct = Content(self.title, 'input', self.var_name,
                     defaultValue=self.default)
        return ct


class Select(AbcArgs):

    def __init__(self, title, source, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

        if isinstance(source, Iterable):
            source = [{'value': i, 'text': str(v)}
                      for i, v in enumerate(source)]
        self.source = source

    def to_content(self) -> Content:
        ct = Content(self.title, 'select', self.var_name,
                     source=self.source, defaultValue=self.default)
        return ct


class Switch(AbcArgs):

    def __init__(self, title, default=None, required=True) -> None:
        super().__init__(title, default=default, required=required)

    def to_content(self) -> Content:
        ct = Content(self.title, 'switch', self.var_name,
                     defaultValue=self.default)
        return ct



from typing import Tuple
from flask import Flask
import flask
from flask import request
import pandas as pd
import numpy as np
import json
import traceback
import sys


from .src.cusFuns.core.FunsPool import auto_register
from .src.helper.utils import json_converter
from .src.dataModel.pandasCmd import PandasCmd
from .src.core.Proxy import ProxyManager

import webbrowser
import os

import pathlib
base_path = pathlib.Path(__file__).parent

m_proxys = ProxyManager()
auto_register(base_path, 'src/cusFuns')

app = Flask(__name__,
            template_folder='./web/templates',
            static_folder='./web/static')


@app.errorhandler(Exception)
def error_handler(ex):
    exc_type, exc_value, exc_traceback_obj = sys.exc_info()
    return {
        'type':
        'error',
        'code':
        traceback.format_exception(exc_type,
                                   exc_value,
                                   exc_traceback_obj,
                                   limit=2),
        'message':
        repr(ex)
    }


def get_data_from_post():
    return json.loads(flask.request.get_data().decode())


@app.route('/')
def index():
    # from . import __version__
    # pd.DataFrame({'a': [1, 2, 3]}).to_csv('test_ret.csv')
    # return f'版本号：{__version__}'

    return flask.render_template('index.html')


@app.route('/api/get_file_args/ext=<ext>', methods=['get'])
def get_file_args(ext: str):
    print(ext)
    print(request.values.get('ext'))
    return 'done'


@app.route('/api/upload_file', methods=['post'])
def upload_file():

    file = flask.request.files.get('file')

    ext = flask.request.form['ext']
    filename = flask.request.form['file_name']
    args = json.loads(flask.request.form['args'])

    # todo:根据不同情况做处理
    m_proxys.read_data(file, filename, ext, args)
    return df2json(m_proxys.get_df_data())


@app.route('/api/query', methods=['post'])
def query():
    param = get_data_from_post()
    m_proxys.query(param['query'])
    return df2json(m_proxys.get_df_data())


@app.route('/api/table/filters', methods=['post'])
def table_filters():
    param = get_data_from_post()
    m_proxys.query(param['query'])
    return df2json(m_proxys.get_df_data())


@app.route('/api/table/handle', methods=['post'])
def table_handle():

    # todo:重构
    def _to_df_cols(df: pd.DataFrame, col: str):
        # 如果是多层索引，使用eval转成元组
        if isinstance(df.columns, pd.MultiIndex):
            return eval(col)

        return col

    param = get_data_from_post()

    filters = param['filters']
    sort = param['sort']

    print(filters, sort)

    filters = ((f['field'], f['values']) for f in filters)
    m_proxys.filter(filters)

    df_ret = m_proxys.get_df_data()

    if sort:
        order = True if sort['order'] == 'asc' else False
        df_ret = df_ret.sort_values(_to_df_cols(
            df_ret, sort['field']), ascending=order)

    return df2json(df_ret)


@app.route('/api/groupby', methods=['post'])
def groupby():
    param = get_data_from_post()
    keys = param['_keys']
    aggs = param['_aggs']

    aggs = {d['filed']: d['methods'] for d in aggs}

    m_proxys.groupby(keys, aggs)

    return df2json(m_proxys.get_df_data())


@app.route('/api/remove/cmd', methods=['post'])
def remove_cmd():
    param = get_data_from_post()
    id = param['id']

    m_proxys.remove_cmd(id)
    return df2json(m_proxys.get_df_data())


@app.route('/api/py_cody')
def get_py_code():
    ret = m_proxys.to_code()
    return ret


@app.route('/api/cmds')
def get_cmds():
    cmds = m_proxys.get_cmds()
    res = flask.json.dumps(cmds, default=json_converter)
    return res


def df2json(df: pd.DataFrame, head_tail_num=10):

    ret = None

    if len(df) > (head_tail_num * 2):
        ret = pd.concat([df.head(head_tail_num), df.tail(head_tail_num)])
    else:
        ret = df

    cmd = PandasCmd.from_df(df, ret)
    ret = flask.json.dumps(cmd, default=json_converter)
    return ret


@app.route('/api/cus_fun', methods=['post'])
def cus_fun():
    param = get_data_from_post()
    data = param['data']

    m_proxys.run_cus_fun(data)

    return df2json(m_proxys.get_df_data())


@app.route('/api/cus_fun/desc')
def get_cus_funcs_desc():
    # auto_register(base_path , 'src/cusFuns')
    ret = m_proxys.get_cus_funcs_desc()
    ret = flask.json.dumps(ret, default=json_converter)
    return ret


@app.route('/api/cus_fun/model', methods=['get'])
def get_cus_funcs_ui_model():

    uid = str(request.args.get('uid'))

    ret = m_proxys.get_ui_model(uid)
    ret = flask.json.dumps(ret, default=json_converter)
    return ret


@app.route('/api/cus_fun/input', methods=['post'])
def get_cus_funcs_input():
    param = get_data_from_post()
    data = param['data']

    m_proxys.input_cus_funcs(data)
    return df2json(m_proxys.get_df_data())


def run():

    port = 5551

    # The reloader has not yet run - open the browser
    if not os.environ.get("WERKZEUG_RUN_MAIN"):
        webbrowser.open_new(f'http://localhost:{port}/')

    # Otherwise, continue as normal
    app.run(host="localhost", port=port)


def find_arg(args, flag, return_value=True):
    try:
        idx = args.index(flag)

        if return_value:
            return True, args[idx+1]
        return idx >= 0, None
    except ValueError:
        pass

    return False, None


def copy_bat():
    import pathlib
    import shutil
    src_path = pathlib.Path(__file__).parent / 'startup_win.bat'
    shutil.copyfile(src_path, 'startup_win.bat')


def main():
    args = sys.argv

    has, dir = find_arg(args, '-dir')
    if has:
        os.chdir(dir)

    has, _ = find_arg(args, '-c', return_value=False)
    if has:
        copy_bat()
        return

    run()


if __name__ == '__main__':
    main()



from typing import List
from .CodeFormatter import CodeFormatter, GetItemCodeFormatter, MethodCodeFormatter, OperatorCodeFormatter

from ..helper.utils import CanJson


class ArgAbc(object):

    def get_runtime_value(self, var_dict):
        raise NotImplementedError


class ArgActor(ArgAbc):

    def __init__(self, org_value, str_value: str) -> None:
        self.org_value = org_value
        self.str_value = str_value

    def get_runtime_value(self, var_dict):
        return self.org_value

    def __str__(self) -> str:
        return CodeFormatter.type2str(self.str_value)


class ArgVar(ArgAbc):

    def __init__(self, var_name) -> None:
        self.var_name = var_name

    def get_runtime_value(self, var_dict):
        return var_dict[self.var_name]

    def __str__(self) -> str:
        return self.var_name


class MethodCall(CanJson):

    def to_json_dict(self):
        # return CanJson.dict_keep_items(self, self.__dict__, ['name', 'args', 'kwargs'])
        return {
            'desc': self.to_code()
        }

    def __init__(self, name) -> None:
        self.name = name
        self.formatter = MethodCodeFormatter()
        self.args = tuple()
        self.kwargs = {}
        self._ex = None

        self._on_after_call_func = lambda call: None

    def on_after_call(self, func):
        self._on_after_call_func = func

    def set_args(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f'name:{self.name};args:{self.args};kws:{self.kwargs}'

    @property
    def has_ex(self) -> bool:
        return self._ex != None

    @property
    def ex(self):
        return self._ex

    def set_exception(self, ex: Exception):
        self._ex = ex

    def to_code(self):
        return self.formatter.to_code(self.name, self.args, self.kwargs)

    def run(self, obj, var_dict):

        args = [
            a if not isinstance(a, ArgAbc) else a.get_runtime_value(var_dict)
            for a in self.args
        ]

        kws = {
            k: (v if not isinstance(v, ArgAbc)
                else v.get_runtime_value(var_dict))
            for k, v in self.kwargs.items()
        }

        return getattr(obj, self.name)(*args, **kws)


class GetItemMethodCall(MethodCall):

    def __init__(self) -> None:
        super().__init__('__getitem__')
        self.formatter = GetItemCodeFormatter()


class OperatorMethodCall(MethodCall):

    def __init__(self, operator, call_name) -> None:
        '''
        operator : '&' , '|' , '==' 
        '''
        super().__init__(call_name)
        self.operator = operator
        self.formatter = OperatorCodeFormatter()

    def to_code(self):
        return self.formatter.to_code(self.operator, self.args, self.kwargs)


class AndOperatorMethodCall(OperatorMethodCall):

    def __init__(self) -> None:
        super().__init__('&', '__and__')


class OrOperatorMethodCall(OperatorMethodCall):

    def __init__(self) -> None:
        super().__init__('|', '__or__')


class EqualOperatorMethodCall(OperatorMethodCall):

    def __init__(self) -> None:
        super().__init__('==', '__eq__')





from typing import Dict, Iterable
import uuid
from enum import Enum
import numpy as np
import pandas as pd


class CanJson(object):
    def to_json_dict(self):
        return self.__dict__

    def dict_keep_items(self, src: Dict, keys: Iterable):
        return {
            k: v
            for k, v in src.items()
            if k in keys
        }


def json_converter(obj):
    if isinstance(obj, CanJson):
        return obj.to_json_dict()

    if isinstance(obj, uuid.UUID):
        return str(obj)

    if isinstance(obj, Enum):
        return obj.value

    if isinstance(obj, pd.Timestamp):
        return str(obj)

    return obj.__dict__





from typing import Callable, Dict, Union
import uuid
import pathlib
import importlib
import inspect
import os


from .DecoratorFuns import CusFunsWrapper, HandleWrapper
from .UIModl import FunModel


class CusFunc(object):
    """
    docstring
    """
    pass

    def __init__(self) -> None:
        self.call_func: Callable = None
        self.uid: str = None
        self.ui_json_func: Callable = None
        self.code_func: Callable = None
        self.fun_name: str = None

        self.reset_uid()

    def reset_uid(self):
        self.uid = str(uuid.uuid1().int)

    def get_ui_json_model(self) -> FunModel:
        fm: FunModel = self.ui_json_func()
        fm.uid = self.uid
        return fm

    def run(self, data):
        """
        docstring
        """
        pass

    @staticmethod
    def _drop_decoreator(code: str) -> str:
        lines = code.split('\n')
        lines = [n for n in lines if n and n[0] != '@']
        return '\n'.join(lines)

    def to_code(self, *args, **kwargs) -> str:
        if self.code_func:
            return self.code_func(*args, **kwargs)

        def_code = inspect.getsource(self.call_func)
        def_code = CusFunc._drop_decoreator(def_code)

        func_name = self.call_func.__name__
        call_code = f'''df = {func_name}(df,*{args},**{kwargs})'''
        return '\n'.join([def_code, call_code])


class FunsPool(object):
    s_instance: 'FunsPool' = None

    @staticmethod
    def get_once() -> 'FunsPool':
        FunsPool.s_instance = FunsPool.s_instance or FunsPool()
        return FunsPool.s_instance

    def __init__(self) -> None:
        self._mapping: Dict[str, CusFunc] = {}

    def register(self, cusFunc: CusFunc):
        self._mapping[cusFunc.uid] = cusFunc

    def get_cusFunc(self, uid: str) -> CusFunc:
        return self._mapping[uid]

    def get_all_desc(self):
        def _to_dict(fm: FunModel):
            return {
                'function_name': fm.function_name,
                'uid': fm.uid
            }

        models = (cf.get_ui_json_model() for cf in self._mapping.values())

        ret = [_to_dict(f)
               for f in models]
        return ret

    def get_ui_model(self, uid: str):
        ret = self.get_cusFunc(uid).get_ui_json_model()
        return ret


def _register_module(module_path: str):
    '''
    module_path : str 
    'src.cusFuns.columnNameChanger.index'
    '''
    pkg = HandleWrapper.__module__.split('.')[0]
    module = importlib.import_module(f'.{module_path}', pkg)
    # [('name',obj)]
    funcs = [(name, func) for name, func in module.__dict__.items()
             if isinstance(func, CusFunsWrapper)]

    cf = CusFunc()

    for name, func in funcs:
        if isinstance(func, HandleWrapper):
            func.auto_create_ui_func()

            cf.call_func = func.handle_func
            cf.ui_json_func = func.ui_json_func
            cf.fun_name = func.fun_name or '未命名操作'
            cf.code_func = func.code_func

    FunsPool.get_once().register(cf)


def auto_register(package_path: Union[str, pathlib.PurePath], base_folder: Union[str, pathlib.PurePath]):
    org_dir = os.getcwd()

    try:
        os.chdir(package_path)
        base_folder = pathlib.Path(base_folder)

        files = (p for p in base_folder.glob('*/*.py')
                 if p.parent.parts[-1] != 'core')

        # print(f'功能池文件：{list(files)}')

        for md_name in files:
            md_name = str(md_name)[:-3]
            md_name = md_name.replace(os.path.sep, '.')

            _register_module(f'{md_name}')
    except Exception:
        raise

    finally:
        os.chdir(org_dir)




# import pandas as pd


# def to_col_type(org_type):
#     mapping = {
#         'object': 'str',
#         'int': 'number',
#         'int64': 'number',
#         'int32': 'number',
#     }

#     stype = str(org_type)

#     return mapping[stype]


# def get_each_field(df: pd.DataFrame):

#     for c, t in zip(df.columns, df.dtypes):
#         field = c
#         topF = c
#         if isinstance(c, tuple):
#             field = '_'.join(c)
#             topF = c[0]

#         yield topF, {
#             'title': f'{field}[{str(t)}]',
#             'field': field, 'type': to_col_type(t),
#             'filters': []
#         }


# def flat_cols(cols):
#     return [
#         '_'.join(c) if isinstance(c, tuple) else c
#         for c in cols
#     ]


# def columns2fields_obj(df: pd.DataFrame):

#     count = 0
#     pre = None
#     ret = []
#     aggFs = []

#     def _append_ret(aggFs, ret, pre_topf):
#         if len(aggFs) > 0:
#             if len(aggFs) == 1:
#                 ret.append(aggFs[0])
#             else:
#                 ret.append({'title': pre_topf, 'children': aggFs})

#     for topf, field in get_each_field(df):

#         if topf != pre:
#             _append_ret(aggFs, ret, pre)
#             aggFs = []

#         aggFs.append(field)
#         pre = topf

#     _append_ret(aggFs, ret, pre)

#     return ret


# if __name__ == "__main__":
#     df = pd.read_excel('data/test.xlsx')
#     ret = df.groupby(['col1']).agg(
#         {'col2': ['first', 'last'], 'col3': ['first']})
#     fields = columns2fields_obj(ret)
#     print(fields)




from typing import Dict, Iterable, List

from ..cusFuns.core.UIModl import CusFunInputRet, FunModel, InputRet

from ..cusFuns.core.FunsPool import FunsPool

from .CodeFormatter import CodeFormatter

from .MethodCall import AndOperatorMethodCall, ArgActor, ArgVar, EqualOperatorMethodCall, MethodCall, GetItemMethodCall, OrOperatorMethodCall
from .Commander import Commander, CommanderManager, Statement
import pandas as pd
import uuid


class Proxy(object):
    def __init__(self) -> None:
        self.cmdManager = CommanderManager()

    def with_cmd(self, name: str = '') -> Commander:
        cmd = Commander(name)
        self.cmdManager.append(cmd)
        return cmd

    def __getattr__(self, name):
        st = self.cmdManager.get_last_cmd().get_last_statement()
        call = MethodCall(name)
        st.addCall(call)
        return st

    def __getitem__(self, key):
        st = self.cmdManager.get_last_cmd().get_last_statement()
        call = GetItemMethodCall()
        st.addCall(call)
        call.set_args(key)
        return st

    def __or__(self, other):
        st = self.cmdManager.get_last_cmd().get_last_statement()
        call = OrOperatorMethodCall()
        st.addCall(call)
        call.set_args(other)
        return st

    def __and__(self, other):
        st = self.cmdManager.get_last_cmd().get_last_statement()
        call = AndOperatorMethodCall()
        st.addCall(call)
        call.set_args(other)
        return st

    def __eq__(self, other):
        st = self.cmdManager.get_last_cmd().get_last_statement()
        call = EqualOperatorMethodCall()
        st.addCall(call)
        call.set_args(other)
        return st

    def to_code(self):
        return self.cmdManager.to_code()


class CallerProxy(Proxy):

    def __init__(self) -> None:
        super().__init__()

    def set_obj_to_var(self, var_name: str, obj):
        self.cmdManager.addVar(var_name, obj)

    def run(self, res_var_name: str):

        obj = None
        var_dict = self.cmdManager.var_dict

        for cmd in self.cmdManager.get_cmds():

            for s in cmd.statements:
                s.run(var_dict)

        if res_var_name:
            return self.cmdManager.getVar(res_var_name)


class ProxyManager(object):

    s_read_func_mapping = {
        'xlsx': 'read_excel',
        'xls': 'read_excel',
        'xlsm': 'read_excel',
        'csv': 'read_csv',
        'feather': 'read_feather',
    }

    s_engine_map = {
        'xls': 'xlrd',
        'xlsx': 'openpyxl',
        'xlsm': 'openpyxl'
    }

    def __init__(self) -> None:
        self.df_proxy = CallerProxy()
        self.pd_read_proxy = CallerProxy()
        self.source_cache: pd.DataFrame = None

    def _get_all_proxys(self) -> Iterable[CallerProxy]:
        yield self.pd_read_proxy
        yield self.df_proxy

    @staticmethod
    def add_read_kws_for_excel(file_ext, kws):
        if file_ext in ProxyManager.s_engine_map:
            kws['engine'] = ProxyManager.s_engine_map[file_ext]

        return kws

    def read_data(self, file, filename, file_ext, read_kws):
        self.df_proxy.cmdManager.clear_all()
        self.pd_read_proxy.cmdManager.clear_all()

        pd_px = self.pd_read_proxy

        with pd_px.with_cmd('导入模块') as cmd:
            cmd.create_code_statement('import pandas as pd')

        with pd_px.with_cmd('加载数据') as cmd:
            ProxyManager.add_read_kws_for_excel(file_ext, read_kws)
            cmd.create_statement('pd', 'df')
            func_name = ProxyManager.s_read_func_mapping[file_ext]
            getattr(pd_px, func_name)(ArgActor(file, filename), **read_kws)

        df = pd_px.run('df')
        self.source_cache = df

        # 表头有任何不是str，转str
        if (df.columns.to_series().map(type) != str).any():
            with self.df_proxy.with_cmd('表头统一转文本') as cmd:
                cmd.create_code_statement(
                    '''cols = [str(c) for c in df.columns]
df.columns=cols''')

    def get_df_data(self) -> pd.DataFrame:
        self.df_proxy.set_obj_to_var('df', self.source_cache)
        ret = self.df_proxy.run('df')
        self.df_proxy.cmdManager.clear_all_var()
        self.df_proxy.set_obj_to_var('pd', pd)
        return ret

    def query(self, query_str: str):
        df_px = self.df_proxy

        with df_px.with_cmd('筛选') as cmd:
            cmd.create_statement('df', 'df')
            df_px.query(query_str)

    def groupby(self, keys, aggs):
        df_px = self.df_proxy

        with df_px.with_cmd('分组统计') as cmd:
            cmd.create_statement('df', 'df')
            df_px.groupby(keys).agg(aggs).reset_index()

    def filter(self, filters):
        """
        filters : [(字段名,值list)]
        [('name',[1,2,3]),('age',[10,20,30])]
        """

        def _get_code(name, values):
            name = CodeFormatter.type2str(name)
            if len(values) > 1:
                return f'(df[{name}].isin({values}))'

            return f'(df[{name}] == {CodeFormatter.type2str(values[0])})'

        df_px = self.df_proxy
        cmd_name = '筛选(表格操作)'
        last_cmd = df_px.cmdManager.get_last_cmd()

        if (last_cmd is not None) and last_cmd.name == cmd_name:
            df_px.cmdManager.remove_cmd(last_cmd._id)

        filters = list(_get_code(name, values) for name, values in filters)
        cond_code = ' & '.join(filters)

        if len(filters) == 0:
            return

        with self.df_proxy.with_cmd(cmd_name) as cmd:
            cmd.create_code_statement(
                '''cond = {cond}
df = df[cond]'''.format(cond=cond_code))

    def run_cus_fun(self, data):
        print(data)

    def _get_all_cmds(self) -> Iterable[Commander]:
        return (
            m
            for p in self._get_all_proxys()
            for m in p.cmdManager.get_cmds()
        )

    def get_cmds(self) -> List[Commander]:
        cmds = list(self._get_all_cmds())
        return cmds

    def get_ui_model(self, uid: str):
        return FunsPool.get_once().get_ui_model(uid)

    def get_cus_funcs_desc(self):
        ret = FunsPool.get_once().get_all_desc()
        return ret

    def input_cus_funcs(self, data: Dict):

        input_ret = CusFunInputRet.from_dict(data)

        uid = input_ret.uid
        cf = FunsPool.get_once().get_cusFunc(uid)

        input_args = {
            c.var_name: c.input
            for c in input_ret.contents
        }

        with self.df_proxy.with_cmd(cf.fun_name) as cmd:
            code = cf.to_code(**input_args)
            cmd.create_code_statement(code)

    def remove_cmd(self, cmd_id: str):
        for p in self._get_all_proxys():
            p.cmdManager.remove_cmd(uuid.UUID(cmd_id))

    def to_code(self) -> str:
        codes = (
            p.to_code()
            for p in self._get_all_proxys()
        )

        ret = '\n\n'.join(codes)
        return ret
